;====================================================================
    include \masm32\include\masm32rt.inc
    .686
    include \masm32\macros\timers.asm
;====================================================================




counter_beginx MACRO loopcount:REQ, priority
        LOCAL label

        IFNDEF __counter__stuff__defined__x
          __counter__stuff__defined__x equ <1>
          .data
          ALIGN 8             ;; Optimal alignment for QWORD
            __counter__qword__count__x  dq 0
            __counter__loop__count__x   dd 0
            __counter__loop__counter__x dd 0
          .code
        ENDIF

        mov __counter__loop__count__x, loopcount
        IFNB <priority>
          invoke GetCurrentProcess
          invoke SetPriorityClass, eax, priority
        ENDIF

        ALIGN 16
      @@:
        ;; Padding to align reference loop label here:
        nops 10

        xor   eax, eax        ;; Use same CPUID input value for each call
        cpuid                 ;; Flush pipe & wait for pending ops to finish
        rdtsc                 ;; Read Time Stamp Counter

        push  edx             ;; Preserve high-order 32 bits of start count
        push  eax             ;; Preserve low-order 32 bits of start count
        mov   __counter__loop__counter__x, loopcount
        xor   eax, eax
        cpuid                 ;; Make sure loop setup instructions finish

        % echo reference loop required bytes = num2str(32-($-@B))

      ;ALIGN 16                ;; Optimal loop alignment for P6
      @@:                     ;; Start an empty reference loop
        sub   __counter__loop__counter__x, 1
        jnz   @B

        xor   eax, eax
        cpuid                 ;; Make sure loop instructions finish
        rdtsc                 ;; Read end count

        ;; Padding to align timing loop label here:
        nops 19

        pop   ecx             ;; Recover low-order 32 bits of start count
        sub   eax, ecx        ;; Low-order 32 bits of overhead count in EAX
        pop   ecx             ;; Recover high-order 32 bits of start count
        sbb   edx, ecx        ;; High-order 32 bits of overhead count in EDX
        push  edx             ;; Preserve high-order 32 bits of overhead count
        push  eax             ;; Preserve low-order 32 bits of overhead count

        xor   eax, eax
        cpuid
        rdtsc
        push  edx             ;; Preserve high-order 32 bits of start count
        push  eax             ;; Preserve low-order 32 bits of start count
        mov   __counter__loop__counter__x, loopcount
        xor   eax, eax
        cpuid                 ;; Make sure loop setup instructions finish

        % echo timing loop required bytes = num2str(64-($-@B))

      ;ALIGN 16                ;; Optimal loop alignment for P6
      label:                  ;; Start test loop
        __counter__loop__label__x equ <label>
    ENDM

    counter_endx MACRO
        sub   __counter__loop__counter__x, 1
        jnz   __counter__loop__label__x

        xor   eax, eax
        cpuid                 ;; Make sure loop instructions finish
        rdtsc                 ;; Read end count
        pop   ecx             ;; Recover low-order 32 bits of start count
        sub   eax, ecx        ;; Low-order 32 bits of test count in EAX
        pop   ecx             ;; Recover high-order 32 bits of start count
        sbb   edx, ecx        ;; High-order 32 bits of test count in EDX
        pop   ecx             ;; Recover low-order 32 bits of overhead count
        sub   eax, ecx        ;; Low-order 32 bits of adjusted count in EAX
        pop   ecx             ;; Recover high-order 32 bits of overhead count
        sbb   edx, ecx        ;; High-order 32 bits of adjusted count in EDX

        mov   DWORD PTR __counter__qword__count__x, eax
        mov   DWORD PTR __counter__qword__count__x + 4, edx

        invoke GetCurrentProcess
        invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

        IFDEF _EMMS
          EMMS
        ENDIF

        finit
        fild  __counter__qword__count__x
        fild  __counter__loop__count__x
        fdiv
        fistp __counter__qword__count__x

        mov   eax, DWORD PTR __counter__qword__count__x
    ENDM

;====================================================================

    ; -------------------------------------------------------
    ; This is an assembly-time random number generator based
    ; on code by George Marsaglia:
    ;   #define znew  ((z=36969*(z&65535)+(z>>16))<<16)
    ;   #define wnew  ((w=18000*(w&65535)+(w>>16))&65535)
    ;   #define MWC   (znew+wnew)
    ; -------------------------------------------------------

    @znew_seed@ = 362436069
    @wnew_seed@ = 521288629

    @rnd MACRO base:REQ
      LOCAL znew, wnew

      @znew_seed@ = 36969 * (@znew_seed@ AND 65535) + (@znew_seed@ SHR 16)
      znew = @znew_seed@ SHL 16

      @wnew_seed@ = 18000 * (@wnew_seed@ AND 65535) + (@wnew_seed@ SHR 16)
      wnew = @wnew_seed@ AND 65535

      EXITM <(znew + wnew) MOD base>
    ENDM

;====================================================================
    .data
        counts dd 128 dup(0)
        xx     dd 0
    .code
;====================================================================
start:
;====================================================================

    xor edi, edi
    invoke Sleep, 4000
    REPEAT 8
        nops @rnd(100)
        REPEAT 8
            counter_begin 1000, REALTIME_PRIORITY_CLASS
                ;adc xx, 123
                ;xchg xx, eax
            counter_end
            mov [counts+edi*4], eax
            inc edi
        ENDM
    ENDM
    REPEAT 8
        nops @rnd(100)
        REPEAT 8
            counter_beginx 1000, REALTIME_PRIORITY_CLASS
                ;adc xx, 123
                ;xchg xx, eax
            counter_endx
            mov [counts+edi*4], eax
            inc edi
        ENDM
    ENDM

    xor edi, edi
    xor ebx, ebx
    .WHILE edi < 64
        print str$([counts+edi*4]),9
        inc ebx
        .IF ebx > 7
            print chr$(13,10)
            xor ebx, ebx
        .ENDIF
        inc edi
    .ENDW
    print chr$(13,10)
    .WHILE edi < 128
        print str$([counts+edi*4]),9
        inc ebx
        .IF ebx > 7
            print chr$(13,10)
            xor ebx, ebx
        .ENDIF
        inc edi
    .ENDW
    print chr$(13,10)

    inkey "Press any key to exit..."
    exit

;====================================================================
end start
